@@ -1,4 +1,4 @@
-/* $Id: client.c 645 2005-09-28 11:30:00Z lathiat $ */
+/* $Id: client.c 1193 2006-04-24 00:18:41Z lennart $ */
 
 /***
   This file is part of avahi.
@@ -39,6 +39,10 @@
 #include "client.h"
 #include "internal.h"
 
+#define AVAHI_CLIENT_DBUS_API_SUPPORTED ((uint32_t) 0x0201)
+
+static int init_server(AvahiClient *client, int *ret_error);
+
 int avahi_client_set_errno (AvahiClient *client, int error) {
     assert(client);
 
@@ -61,15 +65,15 @@ static void client_set_state (AvahiClient *client, AvahiServerState state) {
     client->state = state;
 
     switch (client->state) {
-        case AVAHI_CLIENT_DISCONNECTED:
+        case AVAHI_CLIENT_FAILURE:
             if (client->bus) {
                 dbus_connection_disconnect(client->bus);
                 dbus_connection_unref(client->bus);
                 client->bus = NULL;
             }
-
+            
             /* Fall through */
-
+            
         case AVAHI_CLIENT_S_COLLISION:
         case AVAHI_CLIENT_S_REGISTERING:
 
@@ -83,8 +87,8 @@ static void client_set_state (AvahiClient *client, AvahiServerState state) {
             client->domain_name = NULL;
             break;
 
-        case AVAHI_CLIENT_S_INVALID:
         case AVAHI_CLIENT_S_RUNNING:
+        case AVAHI_CLIENT_CONNECTING:
             break;
             
     }
@@ -100,54 +104,83 @@ static DBusHandlerResult filter_func(DBusConnection *bus, DBusMessage *message,
     assert(bus);
     assert(message);
     
-    dbus_error_init (&error);
+    dbus_error_init(&error);
 
 /*     fprintf(stderr, "dbus: interface=%s, path=%s, member=%s\n", */
 /*             dbus_message_get_interface (message), */
 /*             dbus_message_get_path (message), */
 /*             dbus_message_get_member (message)); */
 
-    if (client->state == AVAHI_CLIENT_DISCONNECTED)
-        goto fail;
-
     if (dbus_message_is_signal(message, DBUS_INTERFACE_LOCAL, "Disconnected")) {
 
         /* The DBUS server died or kicked us */
-        client_set_state(client, AVAHI_CLIENT_DISCONNECTED);
+        avahi_client_set_errno(client, AVAHI_ERR_DISCONNECTED);
+        goto fail;
 
-    } if (dbus_message_is_signal(message, DBUS_INTERFACE_DBUS, "NameOwnerChanged")) {
+    } else if (dbus_message_is_signal(message, DBUS_INTERFACE_DBUS, "NameAcquired")) {
+
+        /* Ignore this message */
+        
+    } else if (dbus_message_is_signal(message, DBUS_INTERFACE_DBUS, "NameOwnerChanged")) {
         char *name, *old, *new;
         
-        if (!(dbus_message_get_args(
+        if (!dbus_message_get_args(
                   message, &error,
                   DBUS_TYPE_STRING, &name,
                   DBUS_TYPE_STRING, &old,
                   DBUS_TYPE_STRING, &new,
-                  DBUS_TYPE_INVALID) || dbus_error_is_set (&error))) {
+                  DBUS_TYPE_INVALID) || dbus_error_is_set(&error)) {
 
             fprintf(stderr, "WARNING: Failed to parse NameOwnerChanged signal: %s\n", error.message);
+            avahi_client_set_errno(client, AVAHI_ERR_DBUS_ERROR);
             goto fail;
         }
 
-        if (strcmp(name, AVAHI_DBUS_NAME) == 0)
+        if (strcmp(name, AVAHI_DBUS_NAME) == 0) {
 
-            /* Regardless if the server lost or acquired its name or
-             * if the name was transfered: our services are no longer
-             * available, so we disconnect ourselves */
-            
-            client_set_state(client, AVAHI_CLIENT_DISCONNECTED);
+            if (avahi_client_is_connected(client)) {
 
+                /* Regardless if the server lost or acquired its name or
+                 * if the name was transfered: our services are no longer
+                 * available, so we disconnect ourselves */
+                avahi_client_set_errno(client, AVAHI_ERR_DISCONNECTED);
+                goto fail;
+                
+            } else if (client->state == AVAHI_CLIENT_CONNECTING && (!old || *old == 0)) {
+                int ret;
+                
+                /* Server appeared */
+                
+                if ((ret = init_server(client, NULL)) < 0) {
+                    avahi_client_set_errno(client, ret);
+                    goto fail;
+                }
+            }
+        }
+
+    } else if (!avahi_client_is_connected(client)) {
+        
+        /* Ignore messages we get in unconnected state */
+        
     } else if (dbus_message_is_signal (message, AVAHI_DBUS_INTERFACE_SERVER, "StateChanged")) {
         int32_t state;
+        char *e = NULL;
+        int c;
         
-        if (!(dbus_message_get_args(
+        if (!dbus_message_get_args(
                   message, &error,
                   DBUS_TYPE_INT32, &state,
-                  DBUS_TYPE_INVALID) || dbus_error_is_set (&error))) {
+                  DBUS_TYPE_STRING, &e,
+                  DBUS_TYPE_INVALID) || dbus_error_is_set (&error)) {
+
             fprintf(stderr, "WARNING: Failed to parse Server.StateChanged signal: %s\n", error.message);
+            avahi_client_set_errno(client, AVAHI_ERR_DBUS_ERROR);
             goto fail;
         }
-            
+
+        if ((c = avahi_error_dbus_to_number(e)) != AVAHI_OK)
+            avahi_client_set_errno(client, c);
+        
         client_set_state(client, (AvahiClientState) state);
 
     } else if (dbus_message_is_signal (message, AVAHI_DBUS_INTERFACE_ENTRY_GROUP, "StateChanged")) {
@@ -161,11 +194,23 @@ static DBusHandlerResult filter_func(DBusConnection *bus, DBusMessage *message,
         
         if (g) {
             int32_t state;
-            if (!(dbus_message_get_args (message, &error, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID)) ||
+            char *e;
+            int c;
+            
+            if (!dbus_message_get_args(
+                      message, &error,
+                      DBUS_TYPE_INT32, &state,
+                      DBUS_TYPE_STRING, &e,
+                      DBUS_TYPE_INVALID) ||
                 dbus_error_is_set(&error)) {
+
                 fprintf(stderr, "WARNING: Failed to parse EntryGroup.StateChanged signal: %s\n", error.message);
+                avahi_client_set_errno(client, AVAHI_ERR_DBUS_ERROR);
                 goto fail;
             }
+
+            if ((c = avahi_error_dbus_to_number(e)) != AVAHI_OK)
+                avahi_client_set_errno(client, c);
             
             avahi_entry_group_set_state(g, state);
         }
@@ -174,46 +219,90 @@ static DBusHandlerResult filter_func(DBusConnection *bus, DBusMessage *message,
         return avahi_domain_browser_event(client, AVAHI_BROWSER_NEW, message);
     else if (dbus_message_is_signal (message, AVAHI_DBUS_INTERFACE_DOMAIN_BROWSER, "ItemRemove")) 
         return avahi_domain_browser_event(client, AVAHI_BROWSER_REMOVE, message);
+    else if (dbus_message_is_signal (message, AVAHI_DBUS_INTERFACE_DOMAIN_BROWSER, "CacheExhausted")) 
+        return avahi_domain_browser_event(client, AVAHI_BROWSER_CACHE_EXHAUSTED, message);
+    else if (dbus_message_is_signal (message, AVAHI_DBUS_INTERFACE_DOMAIN_BROWSER, "AllForNow")) 
+        return avahi_domain_browser_event(client, AVAHI_BROWSER_ALL_FOR_NOW, message);
+    else if (dbus_message_is_signal (message, AVAHI_DBUS_INTERFACE_DOMAIN_BROWSER, "Failure")) 
+        return avahi_domain_browser_event(client, AVAHI_BROWSER_FAILURE, message);
 
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_TYPE_BROWSER, "ItemNew")) 
         return avahi_service_type_browser_event (client, AVAHI_BROWSER_NEW, message);
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_TYPE_BROWSER, "ItemRemove")) 
         return avahi_service_type_browser_event (client, AVAHI_BROWSER_REMOVE, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_TYPE_BROWSER, "CacheExhausted")) 
+        return avahi_service_type_browser_event (client, AVAHI_BROWSER_CACHE_EXHAUSTED, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_TYPE_BROWSER, "AllForNow")) 
+        return avahi_service_type_browser_event (client, AVAHI_BROWSER_ALL_FOR_NOW, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_TYPE_BROWSER, "Failure")) 
+        return avahi_service_type_browser_event (client, AVAHI_BROWSER_FAILURE, message);
 
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_BROWSER, "ItemNew")) 
         return avahi_service_browser_event (client, AVAHI_BROWSER_NEW, message);
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_BROWSER, "ItemRemove")) 
         return avahi_service_browser_event (client, AVAHI_BROWSER_REMOVE, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_BROWSER, "CacheExhausted")) 
+        return avahi_service_browser_event (client, AVAHI_BROWSER_CACHE_EXHAUSTED, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_BROWSER, "AllForNow")) 
+        return avahi_service_browser_event (client, AVAHI_BROWSER_ALL_FOR_NOW, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_BROWSER, "Failure")) 
+        return avahi_service_browser_event (client, AVAHI_BROWSER_FAILURE, message);
 
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_RESOLVER, "Found")) 
         return avahi_service_resolver_event (client, AVAHI_RESOLVER_FOUND, message);
-    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_RESOLVER, "Timeout")) 
-        return avahi_service_resolver_event (client, AVAHI_RESOLVER_TIMEOUT, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_SERVICE_RESOLVER, "Failure")) 
+        return avahi_service_resolver_event (client, AVAHI_RESOLVER_FAILURE, message);
 
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_HOST_NAME_RESOLVER, "Found")) 
         return avahi_host_name_resolver_event (client, AVAHI_RESOLVER_FOUND, message);
-    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_HOST_NAME_RESOLVER, "Timeout")) 
-        return avahi_host_name_resolver_event (client, AVAHI_RESOLVER_TIMEOUT, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_HOST_NAME_RESOLVER, "Failure")) 
+        return avahi_host_name_resolver_event (client, AVAHI_RESOLVER_FAILURE, message);
 
     else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_ADDRESS_RESOLVER, "Found")) 
         return avahi_address_resolver_event (client, AVAHI_RESOLVER_FOUND, message);
-    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_ADDRESS_RESOLVER, "Timeout")) 
-        return avahi_address_resolver_event (client, AVAHI_RESOLVER_TIMEOUT, message);
-    
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_ADDRESS_RESOLVER, "Failure")) 
+        return avahi_address_resolver_event (client, AVAHI_RESOLVER_FAILURE, message);
+
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_RECORD_BROWSER, "ItemNew")) 
+        return avahi_record_browser_event (client, AVAHI_BROWSER_NEW, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_RECORD_BROWSER, "ItemRemove")) 
+        return avahi_record_browser_event (client, AVAHI_BROWSER_REMOVE, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_RECORD_BROWSER, "CacheExhausted")) 
+        return avahi_record_browser_event (client, AVAHI_BROWSER_CACHE_EXHAUSTED, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_RECORD_BROWSER, "AllForNow")) 
+        return avahi_record_browser_event (client, AVAHI_BROWSER_ALL_FOR_NOW, message);
+    else if (dbus_message_is_signal(message, AVAHI_DBUS_INTERFACE_RECORD_BROWSER, "Failure")) 
+        return avahi_record_browser_event (client, AVAHI_BROWSER_FAILURE, message);
+
+    else {
+
+        fprintf(stderr, "WARNING: Unhandled message: interface=%s, path=%s, member=%s\n",
+               dbus_message_get_interface(message), 
+               dbus_message_get_path(message),
+               dbus_message_get_member(message));
+        
+        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+    }
     
     return DBUS_HANDLER_RESULT_HANDLED;
 
 fail:
 
-    dbus_error_free (&error);
+    if (dbus_error_is_set(&error)) {
+        avahi_client_set_errno(client, avahi_error_dbus_to_number(error.name));
+        dbus_error_free(&error);
+    }
+
+    client_set_state(client, AVAHI_CLIENT_FAILURE);
+    
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
 static int get_server_state(AvahiClient *client, int *ret_error) {
-    DBusMessage *message, *reply;
+    DBusMessage *message = NULL, *reply = NULL;
     DBusError error;
     int32_t state;
-    int e;
+    int e = AVAHI_ERR_NO_MEMORY;
     
     assert(client);
 
@@ -223,69 +312,159 @@ static int get_server_state(AvahiClient *client, int *ret_error) {
         goto fail;
 
     reply = dbus_connection_send_with_reply_and_block (client->bus, message, -1, &error);
-    dbus_message_unref(message);
 
-    if (!reply)
+    if (!reply || dbus_error_is_set (&error))
         goto fail;
 
-    if (!(dbus_message_get_args(reply, &error, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID)))
+    if (!dbus_message_get_args(reply, &error, DBUS_TYPE_INT32, &state, DBUS_TYPE_INVALID) ||
+        dbus_error_is_set (&error))
         goto fail;
 
     client_set_state(client, (AvahiServerState) state);
 
+    dbus_message_unref(message);
+    dbus_message_unref(reply);
+    
     return AVAHI_OK;
 
 fail:
     if (dbus_error_is_set(&error)) {
         e = avahi_error_dbus_to_number (error.name);
         dbus_error_free(&error);
-    } else
-        e = AVAHI_ERR_NO_MEMORY;
-
+    } 
+    
     if (ret_error)
         *ret_error = e;
-        
+
+    if (message)
+        dbus_message_unref(message);
+    if (reply)
+        dbus_message_unref(reply);
+    
     return e;
 }
 
-/* This function acts like dbus_bus_get but creates a private
- * connection instead */
-static DBusConnection*
-avahi_dbus_bus_get (DBusError *error)
-{
-    DBusConnection *conn;
-    const char *env_addr;
+static int check_version(AvahiClient *client, int *ret_error) {
+    DBusMessage *message = NULL, *reply  = NULL;
+    DBusError error;
+    uint32_t version;
+    int e = AVAHI_ERR_NO_MEMORY;
+    
+    assert(client);
 
-    env_addr = getenv ("DBUS_SYSTEM_BUS_ADDRESS");
+    dbus_error_init(&error);
 
-    if (env_addr == NULL || (*env_addr == 0))
-    {
-        env_addr = DBUS_SYSTEM_BUS_DEFAULT_ADDRESS;
+    if (!(message = dbus_message_new_method_call(AVAHI_DBUS_NAME, AVAHI_DBUS_PATH_SERVER, AVAHI_DBUS_INTERFACE_SERVER, "GetAPIVersion")))
+        goto fail;
+
+    reply = dbus_connection_send_with_reply_and_block (client->bus, message, -1, &error);
+
+    if (!reply || dbus_error_is_set (&error)) {
+        char *version_str;
+
+        if (!dbus_error_is_set(&error) || strcmp(error.name, DBUS_ERROR_UNKNOWN_METHOD))
+            goto fail;
+
+        /* If the method GetAPIVersion is not known, we look if
+         * GetVersionString matches "avahi 0.6" which is the only
+         * version we support which doesn't have GetAPIVersion() .*/
+        
+        dbus_message_unref(message);
+        if (reply) dbus_message_unref(reply);
+        dbus_error_free(&error);
+
+        if (!(message = dbus_message_new_method_call(AVAHI_DBUS_NAME, AVAHI_DBUS_PATH_SERVER, AVAHI_DBUS_INTERFACE_SERVER, "GetVersionString")))
+            goto fail;
+
+        reply = dbus_connection_send_with_reply_and_block (client->bus, message, -1, &error);
+            
+        if (!reply || dbus_error_is_set (&error))
+            goto fail;
+        
+        if (!dbus_message_get_args (reply, &error, DBUS_TYPE_STRING, &version_str, DBUS_TYPE_INVALID) ||
+            dbus_error_is_set (&error))
+            goto fail;
+
+        version = strcmp(version_str, "avahi 0.6") == 0 ? 0x0201 : 0x0000;
+            
+    } else {
+        
+        if (!dbus_message_get_args (reply, &error, DBUS_TYPE_UINT32, &version, DBUS_TYPE_INVALID) ||
+            dbus_error_is_set(&error))
+            goto fail;
     }
 
-    conn = dbus_connection_open_private (env_addr, error);
-
-    if (!conn)
-        return NULL;
-
-    dbus_connection_set_exit_on_disconnect (conn, FALSE);
-
-    if (!dbus_bus_register (conn, error))
-    {
-        dbus_connection_close (conn);
-        dbus_connection_unref (conn);
-
-        return NULL;
+    /*fprintf(stderr, "API Version 0x%04x\n", version);*/
+    
+    if ((version & 0xFF00) != (AVAHI_CLIENT_DBUS_API_SUPPORTED & 0xFF00) ||
+        (version & 0x00FF) < (AVAHI_CLIENT_DBUS_API_SUPPORTED & 0x00FF)) {
+        e = AVAHI_ERR_VERSION_MISMATCH;
+        goto fail;
     }
 
-    return conn;
+    dbus_message_unref(message);
+    dbus_message_unref(reply);
+               
+    return AVAHI_OK;
+
+fail:
+    if (dbus_error_is_set(&error)) {
+        e = avahi_error_dbus_to_number (error.name);
+        dbus_error_free(&error);
+    } 
+
+    if (ret_error)
+        *ret_error = e;
+
+    if (message)
+        dbus_message_unref(message);
+    if (reply)
+        dbus_message_unref(reply);
+    
+    return e;
 }
 
-AvahiClient *avahi_client_new(const AvahiPoll *poll_api, AvahiClientCallback callback, void *userdata, int *ret_error) {
+static int init_server(AvahiClient *client, int *ret_error) {
+    int r;
+    
+    if ((r = check_version(client, ret_error)) < 0)
+        return r;
+
+    if ((r = get_server_state(client, ret_error)) < 0)
+        return r;
+
+    return AVAHI_OK;
+}
+
+/* This function acts like dbus_bus_get but creates a private
+ * connection instead. Eventually this should be replaced by a DBUS
+ * provided version. */
+static DBusConnection* avahi_dbus_bus_get(DBusError *error) {
+    DBusConnection *c;
+    const char *a;
+
+    if (!(a = getenv("DBUS_SYSTEM_BUS_ADDRESS")) || !*a)
+        a = DBUS_SYSTEM_BUS_DEFAULT_ADDRESS;
+    
+    if (!(c = dbus_connection_open_private(a, error)))
+        return NULL;
+
+    dbus_connection_set_exit_on_disconnect(c, FALSE);
+
+    if (!dbus_bus_register(c, error)) {
+        dbus_connection_close(c);
+        dbus_connection_unref(c);
+        return NULL;
+    }
+
+    return c;
+}
+
+AvahiClient *avahi_client_new(const AvahiPoll *poll_api, AvahiClientFlags flags, AvahiClientCallback callback, void *userdata, int *ret_error) {
     AvahiClient *client = NULL;
     DBusError error;
 
-    dbus_error_init (&error);
+    dbus_error_init(&error);
 
     if (!(client = avahi_new(AvahiClient, 1))) {
         if (ret_error)
@@ -297,8 +476,9 @@ AvahiClient *avahi_client_new(const AvahiPoll *poll_api, AvahiClientCallback cal
     client->error = AVAHI_OK;
     client->callback = callback;
     client->userdata = userdata;
-    client->state = AVAHI_CLIENT_DISCONNECTED;
-
+    client->state = (AvahiClientState) -1;
+    client->flags = flags;
+    
     client->host_name = NULL;
     client->host_name_fqdn = NULL;
     client->domain_name = NULL;
@@ -312,10 +492,13 @@ AvahiClient *avahi_client_new(const AvahiPoll *poll_api, AvahiClientCallback cal
     AVAHI_LLIST_HEAD_INIT(AvahiServiceResolver, client->service_resolvers);
     AVAHI_LLIST_HEAD_INIT(AvahiHostNameResolver, client->host_name_resolvers);
     AVAHI_LLIST_HEAD_INIT(AvahiAddressResolver, client->address_resolvers);
+    AVAHI_LLIST_HEAD_INIT(AvahiRecordBrowser, client->record_browsers);
 
-    if (!(client->bus = avahi_dbus_bus_get(&error)) ||
-        dbus_error_is_set (&error))
+    if (!(client->bus = avahi_dbus_bus_get(&error)) || dbus_error_is_set(&error)) {
+        if (ret_error)
+            *ret_error = AVAHI_ERR_DBUS_ERROR;
         goto fail;
+    }
 
     if (avahi_dbus_connection_glue(client->bus, poll_api) < 0) {
         if (ret_error)
@@ -337,7 +520,7 @@ AvahiClient *avahi_client_new(const AvahiPoll *poll_api, AvahiClientCallback cal
         "path='" AVAHI_DBUS_PATH_SERVER "'",
         &error);
 
-    if (dbus_error_is_set (&error))
+    if (dbus_error_is_set(&error))
         goto fail;
 
     dbus_bus_add_match (
@@ -348,33 +531,43 @@ AvahiClient *avahi_client_new(const AvahiPoll *poll_api, AvahiClientCallback cal
         "path='" DBUS_PATH_DBUS "'",
         &error);
 
-    if (dbus_error_is_set (&error))
+    if (dbus_error_is_set(&error))
         goto fail;
 
-        dbus_bus_add_match (
+    dbus_bus_add_match (
         client->bus,
         "type='signal', "
         "interface='" DBUS_INTERFACE_LOCAL "'",
         &error);
 
-    if (dbus_error_is_set (&error))
+    if (dbus_error_is_set(&error))
         goto fail;
 
+
     if (!(dbus_bus_name_has_owner(client->bus, AVAHI_DBUS_NAME, &error)) ||
         dbus_error_is_set(&error)) {
 
         /* We free the error so its not set, that way the fail target
          * will return the NO_DAEMON error rather than a DBUS error */
-        dbus_error_free (&error);
+        dbus_error_free(&error);
 
-        if (ret_error)
-            *ret_error = AVAHI_ERR_NO_DAEMON;
+        if (!(flags & AVAHI_CLIENT_NO_FAIL)) {
+            
+            if (ret_error)
+                *ret_error = AVAHI_ERR_NO_DAEMON;
         
-        goto fail;
-    }
+            goto fail;
+        }
 
-    if (get_server_state(client, ret_error) < 0)
-        goto fail;
+        /* The user doesn't want this call to fail if the daemon is not
+         * available, so let's return succesfully */
+        client_set_state(client, AVAHI_CLIENT_CONNECTING);
+        
+    } else {
+
+        if (init_server(client, ret_error) < 0)
+            goto fail;
+    }
 
     return client;
 
@@ -385,8 +578,13 @@ fail:
 
     if (dbus_error_is_set(&error)) {
 
-        if (ret_error)
-            *ret_error = avahi_error_dbus_to_number(error.name);
+        if (ret_error) {
+            if (strcmp(error.name, DBUS_ERROR_FILE_NOT_FOUND) == 0)
+                /* DBUS returns this error when the DBUS daemon is not running */
+                *ret_error = AVAHI_ERR_NO_DAEMON;
+            else
+                *ret_error = avahi_error_dbus_to_number(error.name);
+        }
         
         dbus_error_free(&error);
     }
@@ -397,6 +595,11 @@ fail:
 void avahi_client_free(AvahiClient *client) {
     assert(client);
 
+    if (client->bus)
+        /* Disconnect in advance, so that the free() functions won't
+         * issue needless server calls */
+        dbus_connection_disconnect(client->bus);
+    
     while (client->groups)
         avahi_entry_group_free(client->groups);
 
@@ -412,10 +615,17 @@ void avahi_client_free(AvahiClient *client) {
     while (client->service_resolvers)
         avahi_service_resolver_free(client->service_resolvers);
 
-    if (client->bus) {
-        dbus_connection_disconnect(client->bus);
+    while (client->host_name_resolvers)
+        avahi_host_name_resolver_free(client->host_name_resolvers);
+
+    while (client->address_resolvers)
+        avahi_address_resolver_free(client->address_resolvers);
+
+    while (client->record_browsers)
+        avahi_record_browser_free(client->record_browsers);
+    
+    if (client->bus)
         dbus_connection_unref(client->bus);
-    }
 
     avahi_free(client->version_string);
     avahi_free(client->host_name);
@@ -481,10 +691,10 @@ fail:
     return NULL;
 }
 
-const char* avahi_client_get_version_string (AvahiClient *client) {
+const char* avahi_client_get_version_string(AvahiClient *client) {
     assert(client);
 
-    if (client->state == AVAHI_CLIENT_DISCONNECTED) {
+    if (!avahi_client_is_connected(client)) {
         avahi_client_set_errno(client, AVAHI_ERR_BAD_STATE);
         return NULL;
     }
@@ -495,10 +705,10 @@ const char* avahi_client_get_version_string (AvahiClient *client) {
     return client->version_string;
 }
 
-const char* avahi_client_get_domain_name (AvahiClient *client) {
+const char* avahi_client_get_domain_name(AvahiClient *client) {
     assert(client);
 
-    if (client->state == AVAHI_CLIENT_DISCONNECTED) {
+    if (!avahi_client_is_connected(client)) {
         avahi_client_set_errno(client, AVAHI_ERR_BAD_STATE);
         return NULL;
     }
@@ -509,10 +719,10 @@ const char* avahi_client_get_domain_name (AvahiClient *client) {
     return client->domain_name;
 }
 
-const char* avahi_client_get_host_name (AvahiClient *client) {
+const char* avahi_client_get_host_name(AvahiClient *client) {
     assert(client);
     
-    if (client->state == AVAHI_CLIENT_DISCONNECTED) {
+    if (!avahi_client_is_connected(client)) {
         avahi_client_set_errno(client, AVAHI_ERR_BAD_STATE);
         return NULL;
     }
@@ -526,7 +736,7 @@ const char* avahi_client_get_host_name (AvahiClient *client) {
 const char* avahi_client_get_host_name_fqdn (AvahiClient *client) {
     assert(client);
 
-    if (client->state == AVAHI_CLIENT_DISCONNECTED) {
+    if (!avahi_client_is_connected(client)) {
         avahi_client_set_errno(client, AVAHI_ERR_BAD_STATE);
         return NULL;
     }
@@ -604,7 +814,7 @@ uint32_t avahi_client_get_local_service_cookie(AvahiClient *client) {
     DBusError error;
     assert(client);
 
-    if (client->state == AVAHI_CLIENT_DISCONNECTED) {
+    if (!avahi_client_is_connected(client)) {
         avahi_client_set_errno(client, AVAHI_ERR_BAD_STATE);
         return AVAHI_SERVICE_COOKIE_INVALID;
     }
@@ -649,69 +859,11 @@ fail:
     return AVAHI_SERVICE_COOKIE_INVALID;
 }
 
-int avahi_client_is_service_local(AvahiClient *client, AvahiIfIndex interface, AvahiProtocol protocol, const char *name, const char *type, const char *domain) {
-    DBusMessage *message = NULL, *reply = NULL;
-    DBusError error;
-    int32_t i_interface, i_protocol;
-    int b;
-
+int avahi_client_is_connected(AvahiClient *client) {
     assert(client);
-    assert(name);
-    assert(type);
-    assert(domain);
 
-    if (client->state == AVAHI_CLIENT_DISCONNECTED) {
-        avahi_client_set_errno(client, AVAHI_ERR_BAD_STATE);
-        return AVAHI_SERVICE_COOKIE_INVALID;
-    }
-
-    dbus_error_init (&error);
-
-    if (!(message = dbus_message_new_method_call(AVAHI_DBUS_NAME, AVAHI_DBUS_PATH_SERVER, AVAHI_DBUS_INTERFACE_SERVER, "IsServiceLocal"))) {
-        avahi_client_set_errno(client, AVAHI_ERR_NO_MEMORY);
-        goto fail;
-    }
-
-    i_interface = (int32_t) interface;
-    i_protocol = (int32_t) protocol;
-    
-    if (!dbus_message_append_args(
-            message,
-            DBUS_TYPE_INT32, &i_interface,
-            DBUS_TYPE_INT32, &i_protocol,
-            DBUS_TYPE_STRING, &name,
-            DBUS_TYPE_STRING, &type,
-            DBUS_TYPE_STRING, &domain,
-            DBUS_TYPE_INVALID)) {
-        avahi_client_set_errno (client, AVAHI_ERR_NO_MEMORY);
-        goto fail;
-    }
-
-    reply = dbus_connection_send_with_reply_and_block (client->bus, message, -1, &error);
-
-    if (!reply || dbus_error_is_set (&error))
-        goto fail;
-
-    if (!dbus_message_get_args (reply, &error, DBUS_TYPE_BOOLEAN, &b, DBUS_TYPE_INVALID) ||
-        dbus_error_is_set (&error))
-        goto fail;
-    
-    dbus_message_unref(message);
-    dbus_message_unref(reply);
-
-    return b;
-
-fail:
-
-    if (message)
-        dbus_message_unref(message);
-    if (reply)
-        dbus_message_unref(reply);
-    
-    if (dbus_error_is_set(&error)) {
-        avahi_client_set_dbus_error(client, &error);
-        dbus_error_free(&error);
-    }
-
-    return AVAHI_SERVICE_COOKIE_INVALID;
+    return
+        client->bus &&
+        dbus_connection_get_is_connected(client->bus) &&
+        (client->state == AVAHI_CLIENT_S_RUNNING || client->state == AVAHI_CLIENT_S_REGISTERING || client->state == AVAHI_CLIENT_S_COLLISION);
 }
