@@ -1,4 +1,4 @@
-/* $Id: alternative-test.c 332 2005-08-15 14:25:32Z lennart $ */
+/* $Id: alternative-test.c 879 2005-10-26 21:27:40Z lennart $ */
 
 /***
   This file is part of avahi.
@@ -28,7 +28,7 @@
 #include "alternative.h"
 #include "malloc.h"
 
-int main(int argc, char *argv[]) {
+int main(AVAHI_GCC_UNUSED int argc, AVAHI_GCC_UNUSED char *argv[]) {
     char *r = NULL;
     int i, k;
 
