@@ -1,4 +1,4 @@
-/* $Id: dns.c 308 2005-08-13 21:25:09Z lennart $ */
+/* $Id: dns.c 1203 2006-04-24 21:53:18Z lennart $ */
 
 /***
   This file is part of avahi.
@@ -23,13 +23,14 @@
 #include <config.h>
 #endif
 
-#include <netinet/in.h>
-
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
 #include <assert.h>
 
+#include <netinet/in.h>
+
+#include <avahi-common/defs.h>
 #include <avahi-common/domain.h>
 #include <avahi-common/malloc.h>
 
@@ -41,7 +42,7 @@ AvahiDnsPacket* avahi_dns_packet_new(unsigned mtu) {
     size_t max_size;
 
     if (mtu <= 0)
-        max_size = AVAHI_DNS_PACKET_MAX_SIZE;
+        max_size = AVAHI_DNS_PACKET_SIZE_MAX;
     else if (mtu >= AVAHI_DNS_PACKET_EXTRA_SIZE)
         max_size = mtu - AVAHI_DNS_PACKET_EXTRA_SIZE;
     else
@@ -56,6 +57,7 @@ AvahiDnsPacket* avahi_dns_packet_new(unsigned mtu) {
     p->size = p->rindex = AVAHI_DNS_PACKET_HEADER_SIZE;
     p->max_size = max_size;
     p->name_table = NULL;
+    p->data = NULL;
 
     memset(AVAHI_DNS_PACKET_DATA(p), 0, p->size);
     return p;
@@ -165,7 +167,6 @@ uint8_t* avahi_dns_packet_append_name(AvahiDnsPacket *p, const char *name) {
         const char *pname;
         char label[64], *u;
         
-
         /* Check whether we can compress this name. */
 
         if (p->name_table && (prev = avahi_hashmap_lookup(p->name_table, name))) {
@@ -177,11 +178,12 @@ uint8_t* avahi_dns_packet_append_name(AvahiDnsPacket *p, const char *name) {
             assert(idx < p->size);
 
             if (idx < 0x4000) {
-                uint16_t *t;
-                if (!(t = (uint16_t*) avahi_dns_packet_extend(p, sizeof(uint16_t))))
+                uint8_t *t;
+                if (!(t = (uint8_t*) avahi_dns_packet_extend(p, sizeof(uint16_t))))
                     return NULL;
 
-                *t = htons((0xC000 | idx));
+		t[0] = (uint8_t) ((0xC000 | idx) >> 8);
+		t[1] = (uint8_t) idx;
                 return saved_ptr;
             }
         }
@@ -223,7 +225,8 @@ uint8_t* avahi_dns_packet_append_uint16(AvahiDnsPacket *p, uint16_t v) {
     if (!(d = avahi_dns_packet_extend(p, sizeof(uint16_t))))
         return NULL;
     
-    *((uint16_t*) d) = htons(v);
+    d[0] = (uint8_t) (v >> 8);
+    d[1] = (uint8_t) v;
     return d;
 }
 
@@ -234,7 +237,10 @@ uint8_t *avahi_dns_packet_append_uint32(AvahiDnsPacket *p, uint32_t v) {
     if (!(d = avahi_dns_packet_extend(p, sizeof(uint32_t))))
         return NULL;
     
-    *((uint32_t*) d) = htonl(v);
+    d[0] = (uint8_t) (v >> 24);
+    d[1] = (uint8_t) (v >> 16);
+    d[2] = (uint8_t) (v >> 8);
+    d[3] = (uint8_t) v;
 
     return d;
 }
@@ -286,16 +292,31 @@ uint8_t *avahi_dns_packet_extend(AvahiDnsPacket *p, size_t l) {
     return d;
 }
 
-int avahi_dns_packet_is_valid(AvahiDnsPacket *p) {
+int avahi_dns_packet_check_valid(AvahiDnsPacket *p) {
     uint16_t flags;
     assert(p);
 
-    if (p->size < 12)
+    if (p->size < AVAHI_DNS_PACKET_HEADER_SIZE)
         return -1;
 
     flags = avahi_dns_packet_get_field(p, AVAHI_DNS_FIELD_FLAGS);
+    
+    if (flags & AVAHI_DNS_FLAG_OPCODE)
+        return -1;
 
-    if (flags & AVAHI_DNS_FLAG_OPCODE || flags & AVAHI_DNS_FLAG_RCODE)
+    return 0;
+}
+
+int avahi_dns_packet_check_valid_multicast(AvahiDnsPacket *p) {
+    uint16_t flags;
+    assert(p);
+
+    if (avahi_dns_packet_check_valid(p) < 0)
+        return -1;
+    
+    flags = avahi_dns_packet_get_field(p, AVAHI_DNS_FIELD_FLAGS);
+    
+    if (flags & AVAHI_DNS_FLAG_RCODE)
         return -1;
 
     return 0;
@@ -350,7 +371,7 @@ static int consume_labels(AvahiDnsPacket *p, unsigned idx, char *ret_name, size_
             } else
                 first_label = 0;
 
-            if (!(avahi_escape_label(AVAHI_DNS_PACKET_DATA(p) + idx, n, &ret_name, &l)))
+            if (!(avahi_escape_label((char*) AVAHI_DNS_PACKET_DATA(p) + idx, n, &ret_name, &l)))
                 return -1;
 
             idx += n;
@@ -385,26 +406,32 @@ int avahi_dns_packet_consume_name(AvahiDnsPacket *p, char *ret_name, size_t l) {
 }
 
 int avahi_dns_packet_consume_uint16(AvahiDnsPacket *p, uint16_t *ret_v) {
+    uint8_t *d;
+
     assert(p);
     assert(ret_v);
 
     if (p->rindex + sizeof(uint16_t) > p->size)
         return -1;
 
-    *ret_v = ntohs(*((uint16_t*) (AVAHI_DNS_PACKET_DATA(p) + p->rindex)));
+    d = (uint8_t*) (AVAHI_DNS_PACKET_DATA(p) + p->rindex);
+    *ret_v = (d[0] << 8) | d[1];
     p->rindex += sizeof(uint16_t);
 
     return 0;
 }
 
 int avahi_dns_packet_consume_uint32(AvahiDnsPacket *p, uint32_t *ret_v) {
+    uint8_t* d;
+
     assert(p);
     assert(ret_v);
 
     if (p->rindex + sizeof(uint32_t) > p->size)
         return -1;
 
-    *ret_v = ntohl(*((uint32_t*) (AVAHI_DNS_PACKET_DATA(p) + p->rindex)));
+    d = (uint8_t*) (AVAHI_DNS_PACKET_DATA(p) + p->rindex);
+    *ret_v = (d[0] << 24) | (d[1] << 16) | (d[2] << 8) | d[3];
     p->rindex += sizeof(uint32_t);
     
     return 0;
@@ -469,87 +496,59 @@ int avahi_dns_packet_skip(AvahiDnsPacket *p, size_t length) {
     return 0;
 }
 
-AvahiRecord* avahi_dns_packet_consume_record(AvahiDnsPacket *p, int *ret_cache_flush) {
-    char name[257], buf[257];
-    uint16_t type, class;
-    uint32_t ttl;
-    uint16_t rdlength;
-    AvahiRecord *r = NULL;
+static int parse_rdata(AvahiDnsPacket *p, AvahiRecord *r, uint16_t rdlength) {
+    char buf[AVAHI_DOMAIN_NAME_MAX];
     const void* start;
-
+    
     assert(p);
-    assert(ret_cache_flush);
+    assert(r);
 
-/*     avahi_log_debug("consume_record()"); */
-
-    if (avahi_dns_packet_consume_name(p, name, sizeof(name)) < 0 ||
-        avahi_dns_packet_consume_uint16(p, &type) < 0 ||
-        avahi_dns_packet_consume_uint16(p, &class) < 0 ||
-        avahi_dns_packet_consume_uint32(p, &ttl) < 0 ||
-        avahi_dns_packet_consume_uint16(p, &rdlength) < 0 ||
-        p->rindex + rdlength > p->size)
-        goto fail;
-
-/*     avahi_log_debug("name = %s, rdlength = %u", name, rdlength); */
-
-    *ret_cache_flush = !!(class & AVAHI_DNS_CACHE_FLUSH);
-    class &= ~AVAHI_DNS_CACHE_FLUSH;
-    
     start = avahi_dns_packet_get_rptr(p);
-    
-    if (!(r = avahi_record_new_full(name, class, type, ttl)))
-        return NULL;
-    
-    switch (type) {
+ 
+    switch (r->key->type) {
         case AVAHI_DNS_TYPE_PTR:
         case AVAHI_DNS_TYPE_CNAME:
-
-/*             avahi_log_debug("ptr"); */
+        case AVAHI_DNS_TYPE_NS:
             
             if (avahi_dns_packet_consume_name(p, buf, sizeof(buf)) < 0)
-                goto fail;
+                return -1;
 
             r->data.ptr.name = avahi_strdup(buf);
             break;
 
             
         case AVAHI_DNS_TYPE_SRV:
-
-/*             avahi_log_debug("srv"); */
             
             if (avahi_dns_packet_consume_uint16(p, &r->data.srv.priority) < 0 ||
                 avahi_dns_packet_consume_uint16(p, &r->data.srv.weight) < 0 ||
                 avahi_dns_packet_consume_uint16(p, &r->data.srv.port) < 0 ||
                 avahi_dns_packet_consume_name(p, buf, sizeof(buf)) < 0)
-                goto fail;
+                return -1;
             
             r->data.srv.name = avahi_strdup(buf);
             break;
 
         case AVAHI_DNS_TYPE_HINFO:
+
+            if (avahi_dns_packet_consume_string(p, buf, sizeof(buf)) < 0)
+                return -1;
             
-/*             avahi_log_debug("hinfo"); */
-
-            if (avahi_dns_packet_consume_string(p, buf, sizeof(buf)) < 0)
-                goto fail;
-
             r->data.hinfo.cpu = avahi_strdup(buf);
-
+            
             if (avahi_dns_packet_consume_string(p, buf, sizeof(buf)) < 0)
-                goto fail;
-
+                return -1;
+            
             r->data.hinfo.os = avahi_strdup(buf);
             break;
 
         case AVAHI_DNS_TYPE_TXT:
 
-/*             avahi_log_debug("txt"); */
-
             if (rdlength > 0) {
-                r->data.txt.string_list = avahi_string_list_parse(avahi_dns_packet_get_rptr(p), rdlength);
+                if (avahi_string_list_parse(avahi_dns_packet_get_rptr(p), rdlength, &r->data.txt.string_list) < 0)
+                    return -1;
                 
                 if (avahi_dns_packet_skip(p, rdlength) < 0)
-                    goto fail;
+                    return -1;
             } else
                 r->data.txt.string_list = NULL;
             
@@ -560,7 +559,7 @@ AvahiRecord* avahi_dns_packet_consume_record(AvahiDnsPacket *p, int *ret_cache_f
 /*             avahi_log_debug("A"); */
 
             if (avahi_dns_packet_consume_bytes(p, &r->data.a.address, sizeof(AvahiIPv4Address)) < 0)
-                goto fail;
+                return -1;
             
             break;
 
@@ -569,7 +568,7 @@ AvahiRecord* avahi_dns_packet_consume_record(AvahiDnsPacket *p, int *ret_cache_f
 /*             avahi_log_debug("aaaa"); */
             
             if (avahi_dns_packet_consume_bytes(p, &r->data.aaaa.address, sizeof(AvahiIPv6Address)) < 0)
-                goto fail;
+                return -1;
             
             break;
             
@@ -582,18 +581,49 @@ AvahiRecord* avahi_dns_packet_consume_record(AvahiDnsPacket *p, int *ret_cache_f
                 r->data.generic.data = avahi_memdup(avahi_dns_packet_get_rptr(p), rdlength);
                 
                 if (avahi_dns_packet_skip(p, rdlength) < 0)
-                    goto fail;
+                    return -1;
             }
 
             break;
     }
 
-/*     avahi_log_debug("%i == %u ?", (uint8_t*) avahi_dns_packet_get_rptr(p) - (uint8_t*) start, rdlength); */
-    
     /* Check if we read enough data */
     if ((const uint8_t*) avahi_dns_packet_get_rptr(p) - (const uint8_t*) start != rdlength)
+        return -1;
+    
+    return 0;
+}
+
+AvahiRecord* avahi_dns_packet_consume_record(AvahiDnsPacket *p, int *ret_cache_flush) {
+    char name[AVAHI_DOMAIN_NAME_MAX];
+    uint16_t type, class;
+    uint32_t ttl;
+    uint16_t rdlength;
+    AvahiRecord *r = NULL;
+
+    assert(p);
+
+    if (avahi_dns_packet_consume_name(p, name, sizeof(name)) < 0 ||
+        avahi_dns_packet_consume_uint16(p, &type) < 0 ||
+        avahi_dns_packet_consume_uint16(p, &class) < 0 ||
+        avahi_dns_packet_consume_uint32(p, &ttl) < 0 ||
+        avahi_dns_packet_consume_uint16(p, &rdlength) < 0 ||
+        p->rindex + rdlength > p->size)
         goto fail;
 
+    if (ret_cache_flush)
+        *ret_cache_flush = !!(class & AVAHI_DNS_CACHE_FLUSH);
+    class &= ~AVAHI_DNS_CACHE_FLUSH;
+    
+    if (!(r = avahi_record_new_full(name, class, type, ttl)))
+        goto fail;
+    
+    if (parse_rdata(p, r, rdlength) < 0)
+        goto fail;
+
+    if (!avahi_record_is_valid(r))
+        goto fail;
+    
     return r;
 
 fail:
@@ -606,19 +636,29 @@ fail:
 AvahiKey* avahi_dns_packet_consume_key(AvahiDnsPacket *p, int *ret_unicast_response) {
     char name[256];
     uint16_t type, class;
+    AvahiKey *k;
 
     assert(p);
-    assert(ret_unicast_response);
 
     if (avahi_dns_packet_consume_name(p, name, sizeof(name)) < 0 ||
         avahi_dns_packet_consume_uint16(p, &type) < 0 ||
         avahi_dns_packet_consume_uint16(p, &class) < 0)
         return NULL;
 
-    *ret_unicast_response = !!(class & AVAHI_DNS_UNICAST_RESPONSE);
-    class &= ~AVAHI_DNS_UNICAST_RESPONSE;
+    if (ret_unicast_response)
+        *ret_unicast_response = !!(class & AVAHI_DNS_UNICAST_RESPONSE);
 
-    return avahi_key_new(name, class, type);
+    class &= ~AVAHI_DNS_UNICAST_RESPONSE;
+    
+    if (!(k = avahi_key_new(name, class, type)))
+        return NULL;
+
+    if (!avahi_key_is_valid(k)) {
+        avahi_key_unref(k);
+        return NULL;
+    }
+
+    return k;
 }
 
 uint8_t* avahi_dns_packet_append_key(AvahiDnsPacket *p, AvahiKey *k, int unicast_response) {
@@ -640,6 +680,80 @@ uint8_t* avahi_dns_packet_append_key(AvahiDnsPacket *p, AvahiKey *k, int unicast
     return t;
 }
 
+static int append_rdata(AvahiDnsPacket *p, AvahiRecord *r) {
+    assert(p);
+    assert(r);
+    
+    switch (r->key->type) {
+        
+        case AVAHI_DNS_TYPE_PTR:
+        case AVAHI_DNS_TYPE_CNAME:
+        case AVAHI_DNS_TYPE_NS:
+            
+            if (!(avahi_dns_packet_append_name(p, r->data.ptr.name)))
+                return -1;
+            
+            break;
+
+        case AVAHI_DNS_TYPE_SRV:
+
+            if (!avahi_dns_packet_append_uint16(p, r->data.srv.priority) ||
+                !avahi_dns_packet_append_uint16(p, r->data.srv.weight) ||
+                !avahi_dns_packet_append_uint16(p, r->data.srv.port) ||
+                !avahi_dns_packet_append_name(p, r->data.srv.name))
+                return -1;
+
+            break;
+
+        case AVAHI_DNS_TYPE_HINFO:
+            if (!avahi_dns_packet_append_string(p, r->data.hinfo.cpu) ||
+                !avahi_dns_packet_append_string(p, r->data.hinfo.os))
+                return -1;
+
+            break;
+
+        case AVAHI_DNS_TYPE_TXT: {
+
+            uint8_t *data;
+            size_t n;
+
+            n = avahi_string_list_serialize(r->data.txt.string_list, NULL, 0);
+
+            if (!(data = avahi_dns_packet_extend(p, n)))
+                return -1;
+
+            avahi_string_list_serialize(r->data.txt.string_list, data, n);
+            break;
+        }
+
+
+        case AVAHI_DNS_TYPE_A:
+
+            if (!avahi_dns_packet_append_bytes(p, &r->data.a.address, sizeof(r->data.a.address)))
+                return -1;
+            
+            break;
+
+        case AVAHI_DNS_TYPE_AAAA:
+            
+            if (!avahi_dns_packet_append_bytes(p, &r->data.aaaa.address, sizeof(r->data.aaaa.address)))
+                return -1;
+            
+            break;
+            
+        default:
+
+            if (r->data.generic.size)
+                if (avahi_dns_packet_append_bytes(p, r->data.generic.data, r->data.generic.size))
+                    return -1;
+
+            break;
+    }
+
+    return 0;
+}
+
+
 uint8_t* avahi_dns_packet_append_record(AvahiDnsPacket *p, AvahiRecord *r, int cache_flush, unsigned max_ttl) {
     uint8_t *t, *l, *start;
     size_t size;
@@ -658,82 +772,16 @@ uint8_t* avahi_dns_packet_append_record(AvahiDnsPacket *p, AvahiRecord *r, int c
 
     start = avahi_dns_packet_extend(p, 0);
 
-    switch (r->key->type) {
-        
-        case AVAHI_DNS_TYPE_PTR:
-        case AVAHI_DNS_TYPE_CNAME :
-
-            if (!(avahi_dns_packet_append_name(p, r->data.ptr.name)))
-                goto fail;
-            
-            break;
-
-        case AVAHI_DNS_TYPE_SRV:
-
-            if (!avahi_dns_packet_append_uint16(p, r->data.srv.priority) ||
-                !avahi_dns_packet_append_uint16(p, r->data.srv.weight) ||
-                !avahi_dns_packet_append_uint16(p, r->data.srv.port) ||
-                !avahi_dns_packet_append_name(p, r->data.srv.name))
-                goto fail;
-
-            break;
-
-        case AVAHI_DNS_TYPE_HINFO:
-            if (!avahi_dns_packet_append_string(p, r->data.hinfo.cpu) ||
-                !avahi_dns_packet_append_string(p, r->data.hinfo.os))
-                goto fail;
-
-            break;
-
-        case AVAHI_DNS_TYPE_TXT: {
-
-            uint8_t *data;
-            size_t n;
-
-            n = avahi_string_list_serialize(r->data.txt.string_list, NULL, 0);
-
-/*             avahi_log_debug("appending string: %u %p", n, r->data.txt.string_list); */
-
-            if (!(data = avahi_dns_packet_extend(p, n)))
-                goto fail;
-
-            avahi_string_list_serialize(r->data.txt.string_list, data, n);
-            break;
-        }
-
-
-        case AVAHI_DNS_TYPE_A:
-
-            if (!avahi_dns_packet_append_bytes(p, &r->data.a.address, sizeof(r->data.a.address)))
-                goto fail;
-            
-            break;
-
-        case AVAHI_DNS_TYPE_AAAA:
-            
-            if (!avahi_dns_packet_append_bytes(p, &r->data.aaaa.address, sizeof(r->data.aaaa.address)))
-                goto fail;
-            
-            break;
-            
-        default:
-
-            if (r->data.generic.size &&
-                avahi_dns_packet_append_bytes(p, r->data.generic.data, r->data.generic.size))
-                goto fail;
-
-            break;
-    }
-
-
-
+    if (append_rdata(p, r) < 0)
+        goto fail;
     
     size = avahi_dns_packet_extend(p, 0) - start;
     assert(size <= 0xFFFF);
 
 /*     avahi_log_debug("appended %u", size); */
 
-    * (uint16_t*) l = htons((uint16_t) size);
+    l[0] = (uint8_t) ((uint16_t) size >> 8);
+    l[1] = (uint8_t) ((uint16_t) size);
     
     return t;
 
@@ -756,3 +804,46 @@ size_t avahi_dns_packet_space(AvahiDnsPacket *p) {
     
     return p->max_size - p->size;
 }
+
+int avahi_rdata_parse(AvahiRecord *record, const void* rdata, size_t size) {
+    int ret;
+    AvahiDnsPacket p;
+    
+    assert(record);
+    assert(rdata);
+    
+    p.data = (void*) rdata;
+    p.max_size = p.size = size;
+    p.rindex = 0;
+    p.name_table = NULL;
+
+    ret = parse_rdata(&p, record, size);
+
+    assert(!p.name_table);
+    
+    return ret;
+}
+
+size_t avahi_rdata_serialize(AvahiRecord *record, void *rdata, size_t max_size) {
+    int ret;
+    AvahiDnsPacket p;
+    
+    assert(record);
+    assert(rdata);
+    assert(max_size > 0);
+
+    p.data = (void*) rdata;
+    p.max_size = max_size;
+    p.size = p.rindex = 0;
+    p.name_table = NULL;
+
+    ret = append_rdata(&p, record);
+
+    if (p.name_table)
+         avahi_hashmap_free(p.name_table);
+
+    if (ret < 0)
+        return (size_t) -1;
+    
+    return p.size;
+}
