@@ -1,7 +1,7 @@
 #ifndef fooutilhfoo
 #define fooutilhfoo
 
-/* $Id: util.h 306 2005-08-12 21:01:28Z lennart $ */
+/* $Id: util.h 980 2005-11-16 01:15:46Z lennart $ */
 
 /***
   This file is part of avahi.
@@ -30,7 +30,7 @@ AVAHI_C_DECL_BEGIN
 
 void avahi_hexdump(const void *p, size_t size);
 
-char *avahi_format_mac_address(const uint8_t* mac, size_t size);
+char *avahi_format_mac_address(char *t, size_t l, const uint8_t* mac, size_t size);
 
 /** Change every character in the string to upper case (ASCII), return a pointer to the string */
 char *avahi_strup(char *s);
