@@ -1,7 +1,7 @@
 #ifndef foostaticserviceshfoo
 #define foostaticserviceshfoo
 
-/* $Id: static-services.h 115 2005-06-16 22:12:26Z lennart $ */
+/* $Id: static-services.h 907 2005-10-30 18:21:57Z lennart $ */
 
 /***
   This file is part of avahi.
@@ -22,7 +22,7 @@
   USA.
 ***/
 
-void static_service_load(void);
+void static_service_load(int in_chroot);
 void static_service_free_all(void);
 void static_service_add_to_server(void);
 void static_service_remove_from_server(void);
